#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # PIG et JSON et streaming avec les données vélib - correction avec Azure
# 
# Correction.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Récupération des données
# 
# Le code suivant télécharge les données nécessaires [data_velib_paris_2014-11-11_22-23.zip](http://www.xavierdupre.fr/enseignement/complements/data_velib_paris_2014-11-11_22-23.zip).

# In[2]:


import os, datetime
from pyensae.datasource import download_data
if not os.path.exists("velib"):
    os.mkdir("velib")
files=download_data("data_velib_paris_2014-11-11_22-23.zip", website="xdtd", whereTo="velib")
files[:2]


# ## Connexion au cluster et import des données

# In[3]:


import os
blobhp = {}
if "HDCREDENTIALS" in os.environ:
    blobhp["blob_storage"], blobhp["password1"], blobhp["hadoop_server"], blobhp["password2"], blobhp["username"] = \
        os.environ["HDCREDENTIALS"].split("**")
    r = type(blobhp)
else:
    from pyquickhelper.ipythonhelper import open_html_form
    params={"blob_storage":"", "password1":"", "hadoop_server":"", "password2":"", "username":"axavier"}
    r = open_html_form(params=params,title="server + hadoop + credentials", key_save="blobhp")
r


# In[4]:


import pyensae
get_ipython().run_line_magic('load_ext', 'pyensae')
get_ipython().run_line_magic('load_ext', 'pyenbc')
blobstorage = blobhp["blob_storage"]
blobpassword = blobhp["password1"]
hadoop_server = blobhp["hadoop_server"]
hadoop_password = blobhp["password2"]
username = blobhp["username"] + "az"
client, bs =  get_ipython().run_line_magic('hd_open', '')
client, bs


# On uploade les données (sauf si vous l'avez déjà fait une fois) :

# In[5]:


files = [ os.path.join("velib",_) for _ in os.listdir("velib") if ".txt" in _]


# In[6]:


if not client.exists(bs, "hdblobstorage", "velib_1h1/paris.2014-11-11_22-00-18.331391.txt"):
    client.upload(bs, "hdblobstorage", "velib_1h1", files)


# In[7]:


df=get_ipython().run_line_magic('blob_ls', 'hdblobstorage/velib_1h1')
df.head()


# ## Exercice 1 : convertir les valeurs numériques
# 
# Le programme suivant prend comme argument les colonnes à extraire des fichiers textes qui sont enregistrés au format "python". Le streaming sur Azure est sensiblement différent du streaming présenté avec Cloudera. Cette version fonctionne également avec Cloudera. La réciproque n'est pas vraie. Les scripts python sont interprétés avec la machine virtuelle java tout comme pig. La solution suivante s'inspire de [Utilisation de Python avec Hive et Pig dans HDInsight](http://azure.microsoft.com/fr-fr/documentation/articles/hdinsight-python/). Voir également [Writing Jython UDFs](https://help.mortardata.com/technologies/pig/writing_jython_udfs).
# 
# Cette écriture impose de comprendre la façon dont PIG décrit les données et l'utilisation de [schema](http://pig.apache.org/docs/r0.9.1/basic.html#Schemas). Le nom du script doit être **jython.py** pour ce notebook sinon le compilateur PIG ne sait pas dans quel langage interpréter ce script.
# 
# La version de jython utilisée sur le cluster est ``2.5.3 (2.5:c56500f08d34+, Aug 13 2012, 14:54:35) [OpenJDK 64-Bit Server VM (Azul Systems, Inc.)]``.
# 
# **La correction inclut encore un bug mais cela devrait être bientôt corrigé. Cela est dû aux différences Python/Jython.**

# In[8]:


import pyensae


# In[9]:


get_ipython().run_cell_magic('PYTHON', 'jython.py', '\nimport datetime, sys, re\n\n@outputSchema("brow: {(available_bike_stands:int, available_bikes:int, lat:double, lng:double, name:chararray, status:chararray)}")\ndef extract_columns_from_js(row):\n    # pour un programmeur python, les schéma sont contre plutôt intuitifs, \n    # { } veut dire une liste, \n    # ( ) un tuple dont chaque colonne est nommé\n    # j\'écrirai peut-être une fonction détermine le schéma en fonction des données\n        \n    # il faut utiliser des expressions régulières pour découper la ligne\n    # car cette expression ne fonctionne pas sur des lignes trop longues\n    # eval ( row )  --> revient à évaluer une ligne de 500 Ko\n    # Hadoop s\'arrête sans réellement proposer de message d\'erreurs qui mettent sur la bonne voie\n    # dans ces cas-là, il faut relancer le job après avoir commenter des lignes\n    # jusqu\'à trouver celle qui provoque l\'arrêt brutal du programme\n    # arrêt qui ne se vérifie pas en local\n    \n    cols = ["available_bike_stands","available_bikes","lat","lng","name","status"]\n    exp = re.compile ("(\\\\{.*?\\\\})")\n    rec = exp.findall(row)\n    res = []\n    for r in rec :\n        station = eval(r)\n        vals = [ str(station[c]) for c in cols ]\n        res.append(tuple(vals))\n    return res\n')


# La vérification qui suit ne fonctionne que si la fonction à tester prend comme entrée une chaîne de caractères mais rien n'empêche de de créer une telle fonction de façon temporaire juste pour vérifier que le script fonctionne (avec Jython 2.5.3) :

# In[10]:


get_ipython().run_cell_magic('jython', 'jython.py extract_columns_from_js', "[{'address': 'RUE DES CHAMPEAUX (PRES DE LA GARE ROUTIERE) - 93170 BAGNOLET', 'collect_date': datetime.datetime(2014, 11, 11, 22, 2, 18, 47270), 'lng': 2.416170724425901, 'contract_name': 'Paris', 'name': '31705 - CHAMPEAUX (BAGNOLET)', 'banking': 0, 'lat': 48.8645278209514, 'bonus': 0, 'status': 'OPEN', 'available_bikes': 1, 'last_update': datetime.datetime(2014, 11, 11, 21, 55, 22), 'number': 31705, 'available_bike_stands': 49, 'bike_stands': 50}]\n[{'address': 'RUE DES CHAMPEAUX (PRES DE LA GARE ROUTIERE) - 93170 BAGNOLET', 'collect_date': datetime.datetime(2014, 11, 11, 22, 2, 18, 47270), 'lng': 2.416170724425901, 'contract_name': 'Paris', 'name': '31705 - CHAMPEAUX (BAGNOLET)', 'banking': 0, 'lat': 48.8645278209514, 'bonus': 0, 'status': 'OPEN', 'available_bikes': 1, 'last_update': datetime.datetime(2014, 11, 11, 21, 55, 22), 'number': 31705, 'available_bike_stands': 49, 'bike_stands': 50}]\n")


# On supprime la précédente exécution si besoin puis on vérifie que le répertoire contenant les résultats est vide :

# In[11]:


if client.exists(bs, client.account_name, "$PSEUDO/velibpy_results/firstjob"):
    r = client.delete_folder (bs, client.account_name, "$PSEUDO/velibpy_results/firstjob")
    print(r)


# On écrit le script PIG qui utilise plus de colonnes :

# In[12]:


get_ipython().run_cell_magic('PIG', 'json_velib_python.pig', "\nREGISTER '$CONTAINER/$SCRIPTPIG/jython.py' using jython as myfuncs;\n\njspy = LOAD '$CONTAINER/velib_1h1/paris.*.txt' \n       USING PigStorage('\\t') AS (arow:chararray);\n\nDESCRIBE jspy ;\n\nmatrice = FOREACH jspy GENERATE myfuncs.extract_columns_from_js(arow);\nDESCRIBE matrice ;\n\nmultiply = FOREACH matrice GENERATE FLATTEN(brow) ;\nDESCRIBE multiply ;\n\nSTORE multiply INTO '$CONTAINER/$PSEUDO/velibpy_results/firstjob' USING PigStorage('\\t') ;\n")


# On soumet le job :

# In[13]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'json_velib_python.pig -d jython.py -o --stop-on-failure')
jid


# On regarde son statut : 

# In[14]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
(st["id"],st["percentComplete"],st["completed"],
st["status"]["jobComplete"],st["status"]["state"])


# On récupère l'erreur :

# In[15]:


get_ipython().run_line_magic('hd_tail_stderr', 'jid["id"] -n 100')


# In[16]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/velibpy_results')


# On récupère les informations qu'on affiche sous forme de dataframe :

# In[17]:


if os.path.exists("velib_exo1.txt") : os.remove("velib_exo1.txt")
get_ipython().run_line_magic('blob_downmerge', '/$PSEUDO/velibpy_results/firstjob velib_exo1.txt')


# In[18]:


get_ipython().run_line_magic('head', 'velib_exo1.txt')


# In[19]:


import pandas
df = pandas.read_csv("velib_hd.txt", sep="\t",
                     names=["available_bike_stands","available_bikes","lat","lng","name","status"])
df.head()


# ## Exercice 2 : stations fermées
#     
# Les stations fermées ne le sont pas tout le temps. On veut calculer le ratio minutes/stations fermées / total des minutes/stations. Le script python permettant de lire les données ne change pas, il suffit juste de déclarer les sorties numériques comme telles. Quelques détails sur les tables :
# 
# * ``jspy`` : contient les données brutes dans une liste de fichiers
# * ``matrice`` : d'après le job qui précède, la table contient une ligne par stations et par minute, chaque ligne décrit le status de la station
# * ``grstation`` : table ``matrice`` groupée par status
# * ``fermees`` : pour chaque groupe, on aggrégé le nombre de minutes multipliés par le nombre de vélos
# * ``gr*,dist*`` : distribution du nombre de stations (Y) en fonction du nombre de vélos ou places disponibles

# En cas d'exécution précédentes :

# In[20]:


for sub in [ "multiply.txt"]:
    if client.exists(bs, client.account_name, "$PSEUDO/velibpy_results/" + sub):
        r = client.delete_folder (bs, client.account_name, "$PSEUDO/velibpy_results/" + sub)
        print(r)


# On va exécuter le job en deux fois. Le premier job met tout à plat. Le second calcule les aggrégations. La plupart du temps, le travaille de recherche concerne la seconde partie. Mais si le job n'est pas scindé, la première partie est toujours exécutée à chaque itération. Dans ce cas-ci, on scinde le job en deux. La première partie forme une table à partir des données initiales. La seconde les agrègre.

# In[21]:


get_ipython().run_cell_magic('PIG', 'json_velib_python2.pig', "\nREGISTER '$CONTAINER/$SCRIPTPIG/jython.py' using jython as myfuncs;\n\njspy = LOAD '$CONTAINER/velib_1h1/*.txt' USING PigStorage('\\t') AS (arow:chararray);\n\nDESCRIBE jspy ;\n\nmatrice = FOREACH jspy GENERATE myfuncs.extract_columns_from_js(arow);\nDESCRIBE matrice ;\n\nmultiply = FOREACH matrice GENERATE FLATTEN(brow) ;\nDESCRIBE multiply ;\n\nSTORE multiply INTO '$CONTAINER/$PSEUDO/velibpy_results/multiply.txt' USING PigStorage('\\t') ;\n")


# In[22]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'json_velib_python2.pig -d jython.py --stop-on-failure')
jid


# In[23]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
st["id"],st["percentComplete"],st["status"]["jobComplete"]


# In[24]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/velibpy_results/multiply.txt')


# In[25]:


for sub in ["fermees.txt", "distribution_bikes.txt", "distribution_stands.txt"]:
    if client.exists(bs, client.account_name, "$PSEUDO/velibpy_results/" + sub):
        r = client.delete_folder (bs, client.account_name, "$PSEUDO/velibpy_results/" + sub)
        print(r)


# In[26]:


get_ipython().run_cell_magic('PIG', 'json_velib_python3.pig', "\nmultiply = LOAD '$CONTAINER/$PSEUDO/velibpy_results/multiply.txt' USING PigStorage('\\t') AS \n        (available_bike_stands:int, available_bikes:int, lat:double, lng:double, name:chararray, status:chararray) ;\nDESCRIBE multiply ;    \n\ngrstation = GROUP multiply BY status ;  \nDESCRIBE grstation ;\n\nfermees = FOREACH grstation GENERATE\n                 group\n                ,SUM(multiply.available_bikes) AS available_bikes\n                ,SUM(multiply.available_bike_stands) AS available_bike_stands \n                ;\nDESCRIBE fermees ;                \n\ngr_av   = GROUP multiply BY available_bikes ;\nDESCRIBE gr_av;\n\ndist_av = FOREACH gr_av GENERATE group, COUNT(multiply) ;\nDESCRIBE dist_av;\n\ngr_pl   = GROUP multiply BY available_bike_stands ;\nDESCRIBE gr_pl;\n\ndist_pl = FOREACH gr_pl GENERATE group, COUNT(multiply) ;\nDESCRIBE dist_pl;\n\nSTORE fermees INTO '$CONTAINER/$PSEUDO/velibpy_results/fermees.txt' USING PigStorage('\\t') ;\nSTORE dist_av INTO '$CONTAINER/$PSEUDO/velibpy_results/distribution_bikes.txt' USING PigStorage('\\t') ;\nSTORE dist_pl INTO '$CONTAINER/$PSEUDO/velibpy_results/distribution_stands.txt' USING PigStorage('\\t') ;\n")


# In[27]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'json_velib_python3.pig --stop-on-failure')
jid


# In[28]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
st["id"],st["percentComplete"],st["status"]["jobComplete"]


# In[29]:


get_ipython().run_line_magic('tail_stderr', 'jid["id"] 10')


# In[30]:


get_ipython().run_line_magic('blob_ls', '/$PSEUDO/velibpy_results')


# In[31]:


if os.path.exists("distribution_bikes.txt") : os.remove("distribution_bikes.txt")
get_ipython().run_line_magic('blob_downmerge', '/$PSEUDO/velibpy_results/distribution_bikes.txt distribution_bikes.txt')


# In[32]:


import pandas
df = pandas.read_csv("distribution_bikes.txt", sep="\t", names=["nb_velos", "nb_stations_minutes"])
df.head()


# In[33]:


import matplotlib.pyplot as plt
plt.style.use('ggplot')
df.plot(x="nb_velos",y="nb_stations_minutes",kind="bar",figsize=(16,4))


# In[34]:


if os.path.exists("distribution_stands.txt") : os.remove("distribution_stands.txt")
get_ipython().run_line_magic('blob_downmerge', '/$PSEUDO/velibpy_results/distribution_stands.txt distribution_stands.txt')


# In[35]:


df = pandas.read_csv("distribution_stands.txt", sep="\t", names=["nb_places", "nb_stations_minutes"])
df.plot(x="nb_places",y="nb_stations_minutes",kind="bar",figsize=(16,4))


# In[36]:


if os.path.exists("fermees.txt") : os.remove("fermees.txt")
get_ipython().run_line_magic('blob_downmerge', '/$PSEUDO/velibpy_results/fermees.txt fermees.txt')


# In[37]:


df = pandas.read_csv("fermees.txt", sep="\t", names=["status", "nb_velos_stations_minutes", "nb_places_stations_minutes"])
df=df.set_index("status")
df = df.T
df["%close"] = df.CLOSED / (df.CLOSED + df.OPEN)
df


# Ce dernier tableau n'est vrai que dans la mesure où les nombres reportées sont fiables lorsque les stations sont fermées.

# ## Exercice 3 : stations fermées, journée complète
#     
# Appliquer cela à une journée complète revient à lancer le même job sur des données plus grandes. On verra bientôt commencer éviter de recopier le job une seconde fois (introduisant une source potentielle d'erreur).

# In[38]:





# ## Exercice 4 : astuces
# 
# Les erreurs de PIG ne sont pas très explicite surtout si elles se produisent dans le script python. Un moyen simple de débugger est d'attraper les exceptions produites par python et de les récupérer sous PIG (le reste du job est enlevé). On peut tout-à-fait imaginer ajouter la version de python installée sur le cluster ainsi que la liste de modules

# In[39]:


get_ipython().run_cell_magic('PYTHON', 'jython.py', '\nimport sys\n\n@outputSchema("brow:chararray")\ndef information(row):\n    return (";".join([str(sys.version), str(sys.executable)])).replace("\\n"," ")\n')


# On vérifie que le script fonctionne avec jython :

# In[40]:


get_ipython().run_cell_magic('jython', 'jython.py information', "n'importe quoi\n")


# In[41]:


get_ipython().run_cell_magic('PIG', 'info.pig', "\nREGISTER '$CONTAINER/$SCRIPTPIG/jython.py' using jython as myfuncs;\n\njspy = LOAD '$CONTAINER/velib_1h1/*.txt' USING PigStorage('\\t') AS (arow:chararray);\none = LIMIT jspy 1 ;\ninfos  = FOREACH one GENERATE myfuncs.information(arow);\nSTORE infos INTO '$CONTAINER/$PSEUDO/results/infos' USING PigStorage('\\t') ;\n")


# In[42]:


if client.exists(bs, client.account_name, "$PSEUDO/results/infos"):
    r = client.delete_folder (bs, client.account_name, "$PSEUDO/results/infos")
    print(r)


# In[43]:


jid = get_ipython().run_line_magic('hd_pig_submit', 'info.pig jython.py --stop-on-failure')
jid


# In[44]:


st = get_ipython().run_line_magic('hd_job_status', 'jid["id"]')
st["id"],st["percentComplete"],st["status"]["jobComplete"]


# In[45]:


get_ipython().run_line_magic('tail_stderr', 'jid["id"] 10')


# In[46]:


if os.path.exists("infos.txt") : os.remove("infos.txt")
get_ipython().run_line_magic('blob_downmerge', '/$PSEUDO/results/infos infos.txt')


# In[47]:


get_ipython().run_line_magic('head', 'infos.txt')


# In[48]: